"""
Integrated tests for `in3.eth.account` module.
"""
import unittest

import in3
from tests.mock.config import mock_config
from tests.transport import mock_transport


class EthAccountGoerliTestCase(unittest.TestCase):

    def setUp(self):
        # self.client = in3.Client('goerli', in3_config=mock_config)
        self.client = in3.Client('goerli', in3_config=mock_config, transport=mock_transport)

    def test_checksum_address(self):
        missing_0x_address = '1fe2e9bf29AA1938859aF64C413361227d04059A'
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.eth.account.checksum_address(missing_0x_address)
        short_address = '0x1fe2e9bf29AA1938859aF64C413361227d04059'
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.eth.account.checksum_address(short_address)
        mixed_case_address = '0x1fe2e9bf29AA1938859aF64C413361227d04059A'
        address_false = self.client.eth.account.checksum_address(mixed_case_address, False)
        self.assertEqual(address_false, '0x1Fe2E9bf29aa1938859Af64C413361227d04059a')
        address_true = self.client.eth.account.checksum_address(mixed_case_address)
        self.assertNotEqual(address_true, '0x1Fe2E9bf29aa1938859Af64C413361227d04059a')
        # Check lower
        self.assertEqual(self.client.eth.account.checksum_address(mixed_case_address.lower()), address_true)
        self.assertEqual(self.client.eth.account.checksum_address(mixed_case_address.lower(), False), address_false)
        # Check immutable
        self.assertEqual(address_true, self.client.eth.account.checksum_address(address_true))
        self.assertEqual(address_false, self.client.eth.account.checksum_address(address_true, False))

    def test_sign(self):
        message = '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
        private_key = '0xd46e8dd67c5d32be8d46e8dd67c5d32be8058bb8eb970870f072445675058bb8'
        singature = self.client.eth.account.sign(private_key, message)
        self.assertEqual(singature,
                         "0xdd34194276b13f44d3f83401f87da6672dd8dc5905590b7ef44623f424af462f70ce23d37b4" +
                         "ada4ff33f5645df62524402c4fb5cac5dca3bce60331f8d6bc5d41c")

    def test_estimate_gas(self):
        transaction = in3.eth.NewTransaction(
            From="0x132D2A325b8d588cFB9C1188daDdD4d00193E028",
            to="0xF5FEb05CA1b451d60b343f5Ee12df2Cc4ce2691B",
            nonce=5,
            data="0x9a5c90aa0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000" +
                 "000000000000000000000000002386f26fc1000000000000000000000000000015094fc63f03e46fa211eb5708088dda62" +
                 "782dc4000000000000000000000000000000000000000000000004563918244f400000")
        gas = self.client.eth.account.estimate_gas(transaction)
        self.assertGreater(gas, 1000)

    def test_get_transaction_count(self):
        rpc = self.client.eth.account.transaction_count('0x0b56Ae81586D2728Ceaf7C00A6020C5D63f02308')
        self.assertGreaterEqual(rpc, 0)

    def test_get_balance(self):
        result = self.client.eth.account.balance('0x6FA33809667A99A805b610C49EE2042863b1bb83')
        self.assertGreaterEqual(result, 0)

    def test_send_tx(self):
        # 1000000000000000000 == 1 ETH
        # SK to PK 0x0b56Ae81586D2728Ceaf7C00A6020C5D63f02308
        secret = hex(0x9852782BEAD36C64161665586D33391ECEC1CCED7432A1D66FD326D38EA0171F)
        sender = self.client.eth.account.recover(secret)
        receiver = hex(0x6FA33809667A99A805b610C49EE2042863b1bb83)
        tx = in3.eth.NewTransaction(to=receiver, value=1463926659)
        tx_hash = self.client.eth.account.send_transaction(sender, tx)
        self.assertEqual(tx_hash, '0x4456152b5f25509a9f6a4117205700f3b480cd837c855602bce6088a10c2fddd')

    def test_send_raw_transaction(self):
        raw_tx = "0xf867078449504f80825208946fa33809667a99a805b610c49ee2042863b1bb83845741bf83802ea0196e11e25ec9de" + \
                 "57108c7964753a034bb8f1dc02efdd1286b8dd865d21216dafa020ef4407a6d1c5f335ce5685cb49a039f704e32d1393" + \
                 "03576956ca5aaefc7ac3"
        tx_hash = self.client.eth.account.send_raw_transaction(raw_tx)
        self.assertEqual(tx_hash, "0x4456152b5f25509a9f6a4117205700f3b480cd837c855602bce6088a10c2fddd")


class EthAccountKovanTestCase(EthAccountGoerliTestCase):

    def setUp(self):
        # self.client = in3.Client('kovan', in3_config=mock_config)
        self.client = in3.Client('kovan', in3_config=mock_config, transport=mock_transport)

    def test_get_transaction_count(self):
        rpc = self.client.eth.account.transaction_count('0x0b56Ae81586D2728Ceaf7C00A6020C5D63f02308')
        self.assertGreaterEqual(rpc, 0)

    def test_send_tx(self):
        # 1000000000000000000 == 1 ETH
        # SK to PK 0x0b56Ae81586D2728Ceaf7C00A6020C5D63f02308
        secret = hex(0x9852782BEAD36C64161665586D33391ECEC1CCED7432A1D66FD326D38EA0171F)
        sender = self.client.eth.account.recover(secret)
        receiver = hex(0x6FA33809667A99A805b610C49EE2042863b1bb83)
        tx = in3.eth.NewTransaction(to=receiver, value=1463926659)
        tx_hash = self.client.eth.account.send_transaction(sender, tx)
        self.assertEqual(tx_hash, '0x561438bacbd058aca597dd8ebaafbf05df993c83c3224301f33d569c417d0db4')

    def test_send_raw_transaction(self):
        raw_tx = "0xf86780843b9aca00825208946fa33809667a99a805b610c49ee2042863b1bb83845741bf838077a09749142188e173" + \
                 "06d3ddc3f524cc2c503538f3bbea10e831a91fba94d27a809aa04106cb2069e626ff8fec0429a95265dbc26dadd36413" + \
                 "26089fb43569820a7b6d"
        tx_hash = self.client.eth.account.send_raw_transaction(raw_tx)
        self.assertEqual(tx_hash, "0x561438bacbd058aca597dd8ebaafbf05df993c83c3224301f33d569c417d0db4")


class EthAccountTestCase(EthAccountGoerliTestCase):

    def setUp(self):
        # self.client = in3.Client(in3_config=mock_config)
        self.client = in3.Client(in3_config=mock_config, transport=mock_transport)

    def test_get_transaction_count(self):
        rpc = self.client.eth.account.transaction_count('0x6FA33809667A99A805b610C49EE2042863b1bb83')
        self.assertGreater(rpc, 0)

    def test_send_tx(self):
        # 1000000000000000000 == 1 ETH
        # SK to PK 0x0b56Ae81586D2728Ceaf7C00A6020C5D63f02308
        secret = hex(0xAC6D6BF94AD0AC65869EF6A0A47A9F2A201956D4AF3FFBE9DFE679399DACD3D9)
        sender = self.client.eth.account.recover(secret)
        receiver = hex(0x0b56Ae81586D2728Ceaf7C00A6020C5D63f02308)
        tx = in3.eth.NewTransaction(to=receiver, value=1463926659)
        tx_hash = self.client.eth.account.send_transaction(sender, tx)
        self.assertEqual(tx_hash, '0xb13b9d38642216af2545f1b9f882413bcdef13bec21def57c699d3a967d763bc')

    def test_send_raw_transaction(self):
        raw_tx = "0xf868098502540be400825208940b56ae81586d2728ceaf7c00a6020c5d63f02308845741bf838026a070af62c71480" + \
                 "8c7607ebe2cc756f20bed9790e5f5f20b192a83f689157a44bd0a0768a5d94766f1645e652dbef3826eed94074b5481a" + \
                 "453ad12aa52ae6f8e42606"
        tx_hash = self.client.eth.account.send_raw_transaction(raw_tx)
        self.assertEqual(tx_hash, "0xb13b9d38642216af2545f1b9f882413bcdef13bec21def57c699d3a967d763bc")



"""
MVC Pattern Model domain classes for the Incubed client module
"""
import warnings

from in3.eth.model import DataTransferObject, Account


class In3Node(DataTransferObject):
    """
    Registered remote verifier that attest, by signing the block hash, that the requested block and transaction were
    indeed mined are in the correct chain fork.
    Args:
        url (str): Endpoint to post to example: https://in3.slock.it
        index (int): Index within the contract example: 13
        address (in3.Account): Address of the node, which is the public address it iis signing with. example: 0x6C1a01C2aB554930A937B0a2E8105fB47946c679
        deposit (int): Deposit of the node in wei example: 12350000
        props (int): Properties of the node. example: 3
        timeout (int): Time (in seconds) until an owner is able to receive his deposit back after he unregisters himself example: 3600
        registerTime (int): When the node was registered in (unixtime?)
        weight (int): Score based on qualitative metadata to base which nodes to ask signatures from.
    """

    def __init__(self, url: str, address: Account, index: int, deposit: int, props: int, timeout: int,
                 registerTime: int, weight: int):
        self.url = url
        self.index = index
        self.address = address
        self.deposit = deposit
        self.props = props
        self.timeout = timeout
        self.registerTime = registerTime
        self.weight = weight


class NodeList(DataTransferObject):
    """
    List of incubed nodes and its metadata, in3 registry contract from which the list was taken,
    network/registry id, and last block number in the selected chain.
    Args:
        nodes ([In3Node]): list of incubed nodes
        contract (Account): incubed registry contract from which the list was taken
        registryId (str): uuid of this incubed network. one chain could contain more than one incubed networks.
        lastBlockNumber (int): last block signed by the network
        totalServers (int): Total servers number (for integrity?)
    """

    def __init__(self, nodes: [In3Node], contract: Account, registryId: str, lastBlockNumber: int, totalServers: int):
        self.nodes = nodes
        self.contract = contract
        self.registryId = registryId
        self.lastBlockNumber = lastBlockNumber
        self.totalServers = totalServers


class ClientConfig(DataTransferObject):
    """
    Determines the behavior of the in3 client, which chain to connect to and how to manage information security policies.

    Considering integrity is guaranteed and confidentiality is not available on public blockchains, these settings will provide a balance between availability, and financial stake in case of repudiation.

    The newer the block is, higher are the chances it gets repudiated by a fork in the chain. In3 nodes will decide individually to sign on repudiable information, reducing the availability. If the application needs the very latest block, consider using a calculated value in `node_signature_consensus` and set `node_signatures` to zero. This setting is as secure as a light-client.

    The older the block gets, the highest is its availability because of the close-to-zero repudiation risk, but blocks older than circa one year are stored in Archive Nodes, expensive computers, so, despite of the low risk, there are not many nodes available with such information, and they must search for the requested block in its database, lowering the availability as well. If the application needs access to _old_ blocks, consider setting `request_timeout` and `request_retries` to accomodate the time the archive nodes take to fetch the inforamtion.

    The verification policy enforces an extra step of security, adding a financial stake in case of repudiation or false/broken proof. For high security application, consider setting a calculated value in `node_min_deposit` and request as much signatures as necessary in `node_signatures`. Setting `chain_finality_threshold` high will guarantee non-repudiability.

    **All args are Optional. Defaults connect to Ethereum main network with regular security levels.**

    Args:
        chain_finality_threshold (int):  Behavior depends on the chain consensus algorithm: POA - percent of signers needed in order reach finality (% of the validators) i.e.: 60 %. POW - mined blocks on top of the requested, i.e. 8 blocks. Defaults are defined in enum.Chain.
        latest_block_stall (int): Distance considered safe, consensus wise, from the very latest block. Higher values exponentially increases state finality, and therefore data security, as well guaranteeded responses from in3 nodes. example: 10 - will ask for the state from (latestBlock-10).
        account_secret (str): Account SK to sign all in3 requests. (Experimental use `set_account_sk`) example: 0x387a8233c96e1fc0ad5e284353276177af2186e7afa85296f106336e376669f7
        node_signatures (int): Node signatures attesting the response to your request. Will send a separate request for each. example: 3 nodes will have to sign the response.
        node_signature_consensus (int): Useful when signatureCount <= 1. The client will check for consensus in responses. example: 10 - will ask for 10 different nodes and compare results looking for a consensus in the responses.
        node_min_deposit (int): Only nodes owning at least this amount will be chosen to sign responses to your requests. i.e. 1000000000000000000 Wei
        node_list_auto_update (bool): If true the nodelist will be automatically updated. False may compromise data security.
        node_limit (int): Limit nodes stored in the client. example: 150 nodes
        request_timeout (int): Milliseconds before a request times out. example: 100000 ms
        request_retries (int): Maximum times the client will retry to contact a certain node. example: 10 retries
        response_proof_level (str): 'none'|'standard'|'full' Full gets the whole block Patricia-Merkle-Tree, Standard only verifies the specific tree branch concerning the request, None only verifies the root hashes, like a light-client does.
        response_includes_code (bool): If true, every request with the address field will include the data, if existent, that is stored in that wallet/smart-contract. If false, only the code digest is included.
        response_keep_proof (bool): If true, proof data will be kept in every rpc response. False will remove this data after using it to verify the responses. Useful for debugging and manually verifying the proofs.
        cached_blocks (int): Maximum blocks kept in memory. example: 100 last requested blocks
        cached_code_bytes (int): Maximum number of bytes used to cache EVM code in memory. example: 100000 bytes
        in3_registry (dict): In3 Registry Smart Contract configuration data
    """

    def __init__(self,
                 chain_finality_threshold: int = None,
                 account_secret: str = None,
                 latest_block_stall: int = None,
                 node_signatures: int = None,
                 node_signature_consensus: int = None,
                 node_min_deposit: int = None,
                 node_list_auto_update: bool = None,
                 node_limit: int = None,
                 request_timeout: int = None,
                 request_retries: int = None,
                 response_proof_level: str = None,
                 response_includes_code: bool = None,
                 response_keep_proof: bool = None,
                 cached_blocks: int = None,
                 cached_code_bytes: int = None,
                 in3_registry: dict = None):
        self.finality: int = chain_finality_threshold
        self.key: str = account_secret
        self.replaceLatestBlock: int = latest_block_stall
        self.signatureCount: int = node_signatures
        self.requestCount: int = node_signature_consensus
        self.minDeposit: int = node_min_deposit
        self.autoUpdateList: bool = node_list_auto_update
        self.nodeLimit: int = node_limit
        self.timeout: int = request_timeout
        self.maxAttempts: int = request_retries
        self.proof: str = response_proof_level
        self.includeCode: bool = response_includes_code
        self.keepIn3: bool = response_keep_proof
        self.maxBlockCache: int = cached_blocks
        self.maxCodeCache: int = cached_code_bytes
        self.nodes: dict = in3_registry
        if self.key:
            warnings.warn('In3 Config: `account_secret` may cause instability.', DeprecationWarning)


class ChainConfig:
    """
    Default in3 client configuration for each chain see #clientConfig for details.
    """

    def __init__(self, chain_id: int, chain_id_alias: str, client_config: ClientConfig):
        self.chain_id: int = chain_id
        self.chain_id_alias: str = chain_id_alias
        self.client_config: ClientConfig = client_config


chain_configs = {
    "mainnet": ChainConfig(
        chain_id=int(0x1),
        chain_id_alias="mainnet",
        client_config=ClientConfig(
            chain_finality_threshold=10,
            latest_block_stall=10,
            node_signatures=2)
    ),
    "kovan": ChainConfig(
        chain_id=int(0x2a),
        chain_id_alias="kovan",
        client_config=ClientConfig(
            chain_finality_threshold=1,
            latest_block_stall=6,
            node_signatures=1,
            node_signature_consensus=3)
    ),
    "evan": ChainConfig(
        chain_id=int(0x4b1),
        chain_id_alias="evan",
        client_config=ClientConfig(
            chain_finality_threshold=1,
            latest_block_stall=6,
            node_signatures=0,
            node_signature_consensus=5)
    ),
    "goerli": ChainConfig(
        chain_id=int(0x5),
        chain_id_alias="goerli",
        client_config=ClientConfig(
            chain_finality_threshold=1,
            latest_block_stall=6,
            node_signatures=2)
    ),
    "ipfs": ChainConfig(
        chain_id=int(0x7d0),
        chain_id_alias="ipfs",
        client_config=ClientConfig(
            chain_finality_threshold=1,
            latest_block_stall=5,
            node_signatures=1,
            node_signature_consensus=1
        )
    ),
}



"""
Integrated tests for `in3.eth` module.
"""
import unittest

import in3
from tests.mock.config import mock_config
from tests.transport import mock_transport


class EthereumTest(unittest.TestCase):

    def setUp(self):
        # self.client = in3.Client(in3_config=mock_config)
        self.client = in3.Client(in3_config=mock_config, transport=mock_transport)

    def test_ethereum_sha3(self):
        digest = self.client.eth.keccak256('0x68656c6c6f20776f726c64')
        self.assertEqual(digest, '0x47173285a8d7341e5e972fc677286384f802f8ef42a5ec5f03bbfa254cb01fad')

    def test_eth_gasPrice(self):
        self.assertGreater(self.client.eth.gas_price(), 1000000)

    def test_block_number(self):
        result = self.client.eth.block_number()
        self.assertGreater(result, 1000)

    def test_get_block_by_number(self):
        block = self.client.eth.block_by_number(9937218)
        self.assertIsInstance(block, in3.eth.Block)
        block = self.client.eth.block_by_number('latest')
        self.assertIsInstance(block, in3.eth.Block)

    def test_get_transaction_by_hash(self):
        tx_hash = '0xae25a4b673bd87f40ea147a5506cb2ffb38e32ec1efc372c6730a5ba50668ae3'
        tx = self.client.eth.transaction_by_hash(tx_hash)
        self.assertIsInstance(tx, in3.eth.Transaction)
        self.assertEqual(tx_hash, tx.hash)

    def test_get_tx_receipt(self):
        tx_hash = '0xb13b9d38642216af2545f1b9f882413bcdef13bec21def57c699d3a967d763bc'
        result = self.client.eth.transaction_receipt(tx_hash)
        self.assertIsInstance(result, in3.eth.TransactionReceipt)


class EthereumGoerliTest(EthereumTest):

    def setUp(self):
        # self.client = in3.Client('goerli', in3_config=mock_config)
        self.client = in3.Client('goerli', in3_config=mock_config, transport=mock_transport)

    def test_get_block_by_number(self):
        block = self.client.eth.block_by_number(2581719)
        self.assertIsInstance(block, in3.eth.Block)
        block = self.client.eth.block_by_number('latest')
        self.assertIsInstance(block, in3.eth.Block)

    def test_get_transaction_by_hash(self):
        tx_hash = '0x4456152b5f25509a9f6a4117205700f3b480cd837c855602bce6088a10c2fddd'
        tx = self.client.eth.transaction_by_hash(tx_hash)
        self.assertIsInstance(tx, in3.eth.Transaction)
        self.assertEqual(tx_hash, tx.hash)

    def test_get_tx_receipt(self):
        tx_hash = '0x4456152b5f25509a9f6a4117205700f3b480cd837c855602bce6088a10c2fddd'
        result = self.client.eth.transaction_receipt(tx_hash)
        self.assertIsInstance(result, in3.eth.TransactionReceipt)


class EthereumKovanTest(EthereumTest):

    def setUp(self):
        # self.client = in3.Client('kovan', in3_config=mock_config)
        self.client = in3.Client('kovan', in3_config=mock_config, transport=mock_transport)

    def test_get_block_by_number(self):
        block = self.client.eth.block_by_number(18135233)
        self.assertIsInstance(block, in3.eth.Block)
        block = self.client.eth.block_by_number('latest')
        self.assertIsInstance(block, in3.eth.Block)

    def test_get_transaction_by_hash(self):
        tx_hash = '0x561438bacbd058aca597dd8ebaafbf05df993c83c3224301f33d569c417d0db4'
        tx = self.client.eth.transaction_by_hash(tx_hash)
        self.assertIsInstance(tx, in3.eth.Transaction)
        self.assertEqual(tx_hash, tx.hash)

    def test_get_tx_receipt(self):
        tx_hash = '0x561438bacbd058aca597dd8ebaafbf05df993c83c3224301f33d569c417d0db4'
        result = self.client.eth.transaction_receipt(tx_hash)
        self.assertIsInstance(result, in3.eth.TransactionReceipt)


if __name__ == '__main__':
    unittest.main()



"""
Maps libin3 names, types and other definitions
"""
import enum


@enum.unique
class SimpleEnum(enum.Enum):
    """
    Abstract stringify enum class
    """

    def __str__(self):
        return self.value


class EthMethods(SimpleEnum):
    """
    Ethereum API methods
    """
    KECCAK = "web3_sha3"
    ACCOUNTS = "eth_accounts"
    BALANCE = "eth_getBalance"
    STORAGE_AT = "eth_getStorageAt"
    GAS_PRICE = "eth_gasPrice"
    BLOCK_NUMBER = "eth_blockNumber"
    BLOCK_BY_NUMBER = "eth_getBlockByNumber"
    BLOCK_BY_HASH = "eth_getBlockByHash"
    BLOCK_TRANSACTION_COUNT_BY_NUMBER = "eth_getBlockTransactionCountByNumber"
    BLOCK_TRANSACTION_COUNT_BY_HASH = "eth_getBlockTransactionCountByHash"
    TRANSACTION_BY_HASH = "eth_getTransactionByHash"
    TRANSACTION_BY_BLOCKHASH_AND_INDEX = "eth_getTransactionByBlockHashAndIndex"
    TRANSACTION_BY_BLOCKNUMBER_AND_INDEX = "eth_getTransactionByBlockNumberAndIndex"
    TRANSACTION_RECEIPT = "eth_getTransactionReceipt"
    PENDING_TRANSACTIONS = "eth_pendingTransactions"
    TRANSACTION_COUNT = "eth_getTransactionCount"
    SEND_TRANSACTION = "eth_sendTransaction"
    SEND_RAW_TRANSACTION = "eth_sendRawTransaction"
    CALL = "eth_call"
    ESTIMATE_TRANSACTION = "eth_estimateGas"
    UNCLE_COUNT_BY_BLOCK_HASH = "eth_getUncleCountByBlockHash"
    UNCLE_COUNT_BY_BLOCK_NUMBER = "eth_getUncleCountByBlockNumber"
    UNCLE_BY_BLOCKHASH_AND_INDEX = "eth_getUncleByBlockHashAndIndex"
    UNCLE_BY_BLOCKNUMBER_AND_INDEX = "eth_getUncleByBlockNumberAndIndex"
    CODE = "eth_getCode"
    # SIGN = "eth_sign"
    SIGN = "in3_signData"
    NEW_FILTER = "eth_newFilter"
    NEW_BLOCK_FILTER = "eth_newBlockFilter"
    NEW_PENDING_TRANSACTION_FILTER = "eth_newPendingTransactionFilter"
    UNINSTALL_FILTER = "eth_uninstallFilter"
    FILTER_CHANGES = "eth_getFilterChanges"
    FILTER_LOGS = "eth_getFilterLogs"
    LOGS = "eth_getLogs"


class In3Methods(SimpleEnum):
    """
    Incubed API methods
    """
    CONFIGURE = "in3_config"
    ECRECOVER = "in3_ecrecover"
    ENSRESOLVE = "in3_ens"
    SIGN = "in3_sign"
    NODE_STATS = "in3_stats"
    ABI_ENCODE = "in3_abiEncode"
    ABI_DECODE = "in3_abiDecode"
    GET_CONFIG = "in3_getConfig"
    PK_2_PUBLIC = "in3_pk2address"
    PK_2_ADDRESS = "in3_pk2address"
    NODE_LIST = "in3_nodeList"
    CHECKSUM_ADDRESS = "in3_checksumAddress"


class RPCCode(enum.Enum):
    """
    Codes returned from libin3 RPC.
    On success positive values (impl. defined) upto INT_MAX maybe returned
    """
    IN3_OK = 0  # Success
    IN3_EUNKNOWN = -1  # Unknown error - usually accompanied with specific error msg
    IN3_ENOMEM = -2  # No memory
    IN3_ENOTSUP = -3  # Not supported
    IN3_EINVAL = -4  # Invalid value
    IN3_EFIND = -5  # Not found
    IN3_ECONFIG = -6  # Invalid config
    IN3_ELIMIT = -7  # Limit reached
    IN3_EVERS = -8  # Version mismatch
    IN3_EINVALDT = -9  # Data invalid, eg. invalid/incomplete JSON
    IN3_EPASS = -10  # Wrong password
    IN3_ERPC = -11  # RPC error (i.e. in3_ctx_t::error set)
    IN3_ERPCNRES = -12  # RPC no response
    IN3_EUSNURL = -13  # USN URL parse error
    IN3_ETRANS = -14  # Transport error
    IN3_ERANGE = -15  # Not in range
    IN3_WAITING = -16  # the process can not be finished since we are waiting for responses
    IN3_EIGNORE = -17  # Ignorable error
    IN3_EPAYMENT_REQUIRED = -18  # payment required


class BlockAt(SimpleEnum):
    """
    Alias for Ethereum blocks
    """
    EARLIEST = "earliest"
    LATEST = "latest"
    PENDING = "pending"


class In3ProofLevel(SimpleEnum):
    """
    Alias for verification levels.
    Verification is done by calculating Ethereum Trie states requested by the Incubed network ans signed as proofs of a certain state.
    """
    NONE = "none"
    STANDARD = "standard"
    FULL = "full"



"""
Shared Library Runtime module

Loads `libin3` according to host hardware architecture and OS.
Maps symbols, methods and types from the library.
Encapsulates low-level rpc calls into a comprehensive runtime.
"""



from in3.eth.account import EthAccountApi
from in3.eth.contract import EthContractApi
from in3.eth.factory import EthObjectFactory
from in3.eth.model import Transaction, Block, TransactionReceipt
from in3.libin3.enum import EthMethods, BlockAt
from in3.libin3.runtime import In3Runtime


class EthereumApi:
    """
    Module based on Ethereum's api and web3.js
    """

    def __init__(self, runtime: In3Runtime):
        self._runtime = runtime
        self._factory = EthObjectFactory(runtime)
        self.account = EthAccountApi(runtime, self._factory)
        self.contract = EthContractApi(runtime, self._factory)

    def keccak256(self, message: str) -> str:
        """
        Keccak-256 digest of the given data. Compatible with Ethereum but not with SHA3-256.
        Args:
            message (str): Message to be hashed.
        Returns:
            digest (str): The message digest.
        """
        return self._runtime.call(EthMethods.KECCAK, message)

    def gas_price(self) -> int:
        """
        The current gas price in Wei (1 ETH equals 1000000000000000000 Wei ).
        Returns:
            price (int): minimum gas value for the transaction to be mined
        """
        return self._factory.get_integer(self._runtime.call(EthMethods.GAS_PRICE))

    def block_number(self) -> int:
        """
        Returns the number of the most recent block the in3 network can collect signatures to verify.
        Can be changed by Client.Config.replaceLatestBlock.
        If you need the very latest block, change Client.Config.signatureCount to zero.
        Returns:
            block_number (int) : Number of the most recent block
        """
        return self._factory.get_integer(self._runtime.call(EthMethods.BLOCK_NUMBER))

    def block_by_hash(self, block_hash: str, get_full_block: bool = False) -> BlockAt:
        """
        Blocks can be identified by root hash of the block merkle tree (this), or sequential number in which it was mined (get_block_by_number).
        Args:
            block_hash (str):  Desired block hash
            get_full_block (bool): If true, returns the full transaction objects, otherwise only its hashes.
        Returns:
            block (Block): Desired block, if exists.
        """
        serialized: dict = self._runtime.call(EthMethods.BLOCK_BY_HASH, self._factory.get_hash(block_hash),
                                              get_full_block)
        return self._factory.get_block(serialized)

    def block_by_number(self, block_number: [int or str], get_full_block: bool = False) -> Block:
        """
        Blocks can be identified by sequential number in which it was mined, or root hash of the block merkle tree (this) (get_block_by_hash).
        Args:
            block_number (int or str):  Desired block number integer or 'latest', 'earliest', 'pending'.
            get_full_block (bool): If true, returns the full transaction objects, otherwise only its hashes.
        Returns:
            block (Block): Desired block, if exists.
        """
        if isinstance(block_number, str) and not block_number.lower() in [str(e) for e in BlockAt]:
            raise AssertionError('Block number must be an integer.')
        block_number_str = hex(block_number) if isinstance(block_number, int) else block_number
        serialized: dict = self._runtime.call(EthMethods.BLOCK_BY_NUMBER, block_number_str, get_full_block)
        return self._factory.get_block(serialized)

    def transaction_by_hash(self, tx_hash: str) -> Transaction:
        """
        Transactions can be identified by root hash of the transaction merkle tree (this) or by its position in the block transactions merkle tree.
        Every transaction hash is unique for the whole chain. Collision could in theory happen, chances are 67148E-63%.
        Args:
            tx_hash: Transaction hash.
        Returns:
            transaction: Desired transaction, if exists.
        """
        serialized: dict = self._runtime.call(EthMethods.TRANSACTION_BY_HASH, self._factory.get_hash(tx_hash))
        return self._factory.get_transaction(serialized)

    def transaction_receipt(self, tx_hash: str) -> TransactionReceipt:
        """
        After a transaction is received the by the client, it returns the transaction hash. With it, it is possible to
        gather the receipt, once a miner has mined and it is part of an acknowledged block. Because how it is possible,
        in distributed systems, that data is asymmetric in different parts of the system, the transaction is only "final"
        once a certain number of blocks was mined after it, and still it can be possible that the transaction is discarded
        after some time. But, in general terms, it is accepted that after 6 to 8 blocks from latest, that it is very
        likely that the transaction will stay in the chain.
        Args:
            tx_hash: Transaction hash.
        Returns:
            tx_receipt: The mined Transaction data including event logs.
        """
        tx_receipt = self._runtime.execute(EthMethods.TRANSACTION_RECEIPT, self._factory.get_hash(tx_hash))
        return self._factory.get_tx_receipt(tx_receipt)



"""
Configuration for `in3.client.Client`. Enables mocked in3 network responses to be valid.
Disable node list updates, and collect signed responses from both boot nodes.
"""
from in3 import ClientConfig

_registry = {
    '0x1': {
        'contract': '0xac1b824795e1eb1f6e609fe0da9b9af8beaab60f',
        'registryId': '0x23d5345c5c13180a8080bd5ddbe7cde64683755dcce6e734d95b7b573845facb',
        'needsUpdate': False,
        'avgBlockTime': 15},
    '0x5': {
        'contract': '0x5f51e413581dd76759e9eed51e63d14c8d1379c8',
        'registryId': '0x67c02e5e272f9d6b4a33716614061dd298283f86351079ef903bf0d4410a44ea',
        'needsUpdate': False,
        'avgBlockTime': 15},
    '0x2a': {
        'contract': '0x4c396dcf50ac396e5fdea18163251699b5fcca25',
        'registryId': '0x92eb6ad5ed9068a24c1c85276cd7eb11eda1e8c50b17fbaffaf3e8396df4becf',
        'needsUpdate': False,
        'avgBlockTime': 6},
}

mock_config = ClientConfig(node_list_auto_update=False,
                           node_signature_consensus=2,
                           node_signatures=2,
                           request_retries=1,
                           in3_registry=_registry,
                           cached_blocks=0,
                           cached_code_bytes=0)



"""
Encapsulates low-level rpc calls into a comprehensive runtime.
"""
import ctypes as c
import json
from enum import Enum

from in3.exception import ClientException
from in3.libin3.enum import RPCCode
from in3.libin3.lib_loader import libin3_new, libin3_free, libin3_call, libin3_exec, libin3_set_pk


class RPCExecRequest:
    """
    RPC request to libin3
    Args:
        fn_name: Name of function that will be called in libin3
        args: Arguments matching the parameters order of this function
        _id: (optional) Message id
        rpc_version: (optional) Version of the json-rpc api
    """

    def __init__(self, fn_name: str, args: tuple = (), verification: str = "proof", _id: int = 1,
                 rpc_version: str = "2.0"):
        self.id = _id
        self.rpc_version = rpc_version
        self.libin3_function = fn_name
        self.args = args
        self.in3_headers = {
            "verification": verification,
            "version": "2.1.0"
        }

    def __bytes__(self):
        """
        [{"id":1,"jsonrpc":"2.0","method":"in3_nodeList","params":[],"in3":{"verification":"proof","version": "2.1.0"}}]
        """
        return json.dumps({
            "id": self.id,
            "jsonrpc": self.rpc_version,
            "method": self.libin3_function,
            "params": [arg for arg in self.args],
            "in3": self.in3_headers
        }).encode('utf8')


class RPCExecResponse:
    """
    RPC response from libin3.
    Args:
        id: Message id
        jsonrpc: Version of the json-rpc api
        result: Function returned value(s)
        error: (Optional) Filled in case of error
    """

    def __init__(self, id: str, jsonrpc: str, result: str = None, error: str = None):
        self.id = id
        self.jsonrpc = jsonrpc
        self.result = result
        self.error = error


class RPCCallRequest:
    """
    RPC request to libin3
    Args:
        fn_name: Name of function that will be called in libin3
        fn_args: Arguments matching the parameters order of this function
    """

    def __init__(self, fn_name: str or Enum, fn_args: tuple = None, formatted: bool = False):
        self.fn_name = str(fn_name).encode('utf8')
        if fn_args and formatted:
            self.fn_args = fn_args[0].encode('utf8')
        elif fn_args and not formatted:
            self.fn_args = json.dumps(list(fn_args)).replace('\'', '').encode('utf8')
        else:
            self.fn_args = '[]'.encode('utf8')


class In3Runtime:
    """
    Instantiate libin3 and frees it when garbage collected.
    Args:
        chain_id (int): Chain-id based on EIP-155. If None provided, will connect to the Ethereum network. i.e: 0x1 for mainNet
    """

    def __init__(self, chain_id: int, transport: c.CFUNCTYPE):
        self.in3 = libin3_new(chain_id, transport)
        self.chain_id = chain_id

    def __del__(self):
        libin3_free(self.in3)

    def call(self, fn_name: str or Enum, *fn_args, formatted: bool = False) -> str or dict:
        """
        Make a remote procedure call to a function in libin3
        Args:
            fn_name (str or Enum): Name of the function to be called
            fn_args: Arguments matching the parameters order of this function
            formatted (bool): True if args must be sent as-is to RPC endpoint
        Returns:
            fn_return (str): String of values returned by the function, if any.
        """
        request = RPCCallRequest(fn_name, fn_args, formatted)
        result, response, error = libin3_call(self.in3, request.fn_name, request.fn_args)
        in3_code = RPCCode(result)
        if not in3_code == RPCCode.IN3_OK or error:
            raise ClientException(str(error))
        return json.loads(response)

    def set_signer_account(self, secret: int) -> int:
        """
        Load an account secret to sign Ethereum transactions with `eth_sendTransaction` and `eth_call`.
        Args:
            secret: SK 256 bit number. example: int(0x387a8233c96e1fc0ad5e284353276177af2186e7afa85296f106336e376669f7, 16)
        """
        return libin3_set_pk(self.in3, hex(secret).encode('utf8'))

    def execute(self, fn_name: str or Enum, *args) -> str or dict:
        """
        Make a remote procedure call to a function in libin3
        Args:
            fn_name (str or Enum): Name of the function to be called
            *args: Arguments matching the parameters order of this function
        Returns:
            fn_return (str): String of values returned by the function, if any.
        """
        request = RPCExecRequest(str(fn_name), args=args)
        request_bytes = bytes(request)
        response_bytes = libin3_exec(self.in3, rpc=request_bytes)
        response_str = response_bytes.decode('utf8').replace('\n', ' ')
        if 'error' in response_str:
            raise ClientException(response_str)
        response_dict = json.loads(response_str)
        response = RPCExecResponse(**response_dict)
        return response.result



import json


class DataTransferObject:
    """
    Maps marshalling objects transferred to, and from a remote facade, in this case, libin3 rpc api.
    For more on design-patterns see [Martin Fowler's](https://martinfowler.com/eaaCatalog/) Catalog of Patterns of Enterprise Application Architecture.
    """

    def to_dict(self, int_to_hex: bool = False) -> dict:
        dictionary = {k: v for k, v in self.__dict__.items() if v is not None}
        if int_to_hex:
            integers = {k: hex(v) for k, v in dictionary.items() if isinstance(v, int)}
            dictionary.update(integers)
        return dictionary

    def serialize(self) -> str:
        return json.dumps(self.to_dict())


class Transaction(DataTransferObject):
    """
    Args:
        From (hex str): Address of the sender account.
        to (hex str): Address of the receiver account. Left undefined for a contract creation transaction.
        gas (int): Gas for the transaction miners and execution in wei. Will get multiplied by `gasPrice`. Use in3.eth.account.estimate_gas to get a calculated value. Set too low and the transaction will run out of gas.
        value (int): Value transferred in wei. The endowment for a contract creation transaction.
        data (hex str): Either a ABI byte string containing the data of the function call on a contract, or in the case of a contract-creation transaction the initialisation code.
        gasPrice (int): Price of gas in wei, defaults to in3.eth.gasPrice. Also know as `tx fee price`. Set your gas price too low and your transaction may get stuck. Set too high on your own loss.
        gasLimit (int); Maximum gas paid for this transaction. Set by the client using this rationale if left empty: gasLimit = G(transaction) + G(txdatanonzero) × dataByteLength. Minimum is 21000.
        nonce (int): Number of transactions mined from this address. Nonce is a value that will make a transaction fail in case it is different from (transaction count + 1). It exists to mitigate replay attacks. This allows to overwrite your own pending transactions by sending a new one with the same nonce. Use in3.eth.account.get_transaction_count to get the latest value.
        hash (hex str): Keccak of the transaction bytes, not part of the transaction. Also known as receipt, because this field is filled after the transaction is sent, by eth_sendTransaction
        blockHash (hex str): Block hash that this transaction was mined in. null when its pending.
        blockHash (int): Block number that this transaction was mined in. null when its pending.
        transactionIndex (int): Integer of the transactions index position in the block. null when its pending.
        signature (hex str): ECDSA of transaction.data, calculated r, s and v concatenated. V is parity set by v = 27 + (r % 2).
    """

    def __init__(self, From: str, to: str, gas: int, gasPrice: int, hash: str, nonce: int, transactionIndex: int,
                 blockHash: str, value: int, input: str, publicKey: str, standardV: int, raw: str, creates: str,
                 chainId: int, r: int, s: int, v: int):
        self.blockHash = blockHash
        self.From = From
        self.gas = gas
        self.gasPrice = gasPrice
        self.hash = hash
        self.nonce = nonce
        self.to = to
        self.transactionIndex = transactionIndex
        self.value = value
        self.input = input
        self.publicKey = publicKey
        self.raw = raw
        self.standardV = standardV
        self.creates = creates
        self.chainId = chainId
        self.r = r
        self.s = s
        self.v = v


class NewTransaction(DataTransferObject):
    """
    Unsent transaction. Use to send a new transaction.
    Args:
        From (hex str): Address of the sender account.
        to (hex str): Address of the receiver account. Left undefined for a contract creation transaction.
        value (int): (optional) Value transferred in wei. The endowment for a contract creation transaction.
        data (hex str): (optional) Either a ABI byte string containing the data of the function call on a contract, or in the case of a contract-creation transaction the initialisation code.
        gasPrice (int): (optional) Price of gas in wei, defaults to in3.eth.gasPrice. Also know as `tx fee price`. Set your gas price too low and your transaction may get stuck. Set too high on your own loss.
        gasLimit (int); (optional) Maximum gas paid for this transaction. Set by the client using this rationale if left empty: gasLimit = G(transaction) + G(txdatanonzero) × dataByteLength. Minimum is 21000.
        nonce (int): (optional) Number of transactions mined from this address. Nonce is a value that will make a transaction fail in case it is different from (transaction count + 1). It exists to mitigate replay attacks. This allows to overwrite your own pending transactions by sending a new one with the same nonce. Use in3.eth.account.get_transaction_count to get the latest value.
        hash (hex str): (optional) Keccak of the transaction bytes, not part of the transaction. Also known as receipt, because this field is filled after the transaction is sent.
        signature (hex str): (optional) ECDSA of transaction, r, s and v concatenated. V is parity set by v = 27 + (r % 2).
    """

    def __init__(self, From: str = None, to: str = None, nonce: int = None, value: int = None,
                 data: str = None, gasPrice: int = None, gasLimit: int = None, hash: str = None, signature: str = None):
        self.From = From
        self.gasPrice = gasPrice
        self.gasLimit = gasLimit
        self.hash = hash
        self.data = data
        self.nonce = nonce
        self.to = to
        self.value = value
        self.signature = signature
        super().__init__()

    def serialize(self) -> dict:
        dictionary = self.to_dict(int_to_hex=True)
        if 'From' in dictionary:
            del dictionary['From']
            dictionary['from'] = self.From
        return dictionary


class Block(DataTransferObject):

    def __init__(self, number: int, hash: str, parentHash: str, nonce: int, sha3Uncles: list, logsBloom: str,
                 transactionsRoot: str, stateRoot: str, miner: str, difficulty: int, totalDifficulty: int,
                 extraData: str, size: int, gasLimit: int, gasUsed: int, timestamp: int, transactions: list,
                 uncles: list, author: str):
        self.number = number
        self.hash = hash
        self.parentHash = parentHash
        self.nonce = nonce
        self.sha3Uncles = sha3Uncles
        self.logsBloom = logsBloom
        self.transactionsRoot = transactionsRoot
        self.stateRoot = stateRoot
        self.miner = miner
        self.difficulty = difficulty
        self.totalDifficulty = totalDifficulty
        self.extraData = extraData
        self.size = size
        self.gasLimit = gasLimit
        self.gasUsed = gasUsed
        self.timestamp = timestamp
        self.transactions = transactions
        self.uncles = uncles
        self.author = author


class Filter(DataTransferObject):
    """
    Filters are event catchers running on the Ethereum Client. Incubed has a client-side implementation.
    An event will be stored in case it is within to and from blocks, or in the block of blockhash, contains a
    transaction to the designed address, and has a word listed on topics.
    """

    def __init__(self, fromBlock: int or str, toBlock: int or str, address: str, topics: list, blockhash: str):
        self.fromBlock = fromBlock
        self.toBlock = toBlock
        self.address = address
        self.topics = topics
        self.blockhash = blockhash


class Log(DataTransferObject):
    """
    Transaction Log for events and data returned from smart-contract method calls.
    """

    def __init__(self, address: hex, blockHash: hex, blockNumber: int, data: hex, logIndex: int, removed: bool,
                 topics: [hex], transactionHash: hex, transactionIndex: int, transactionLogIndex: int, Type: str):
        self.address = address
        self.blockHash = blockHash
        self.blockNumber = blockNumber
        self.data = data
        self.logIndex = logIndex
        self.removed = removed
        self.topics = topics
        self.transactionHash = transactionHash
        self.transactionIndex = transactionIndex
        self.transactionLogIndex = transactionLogIndex
        self.Type = Type


class TransactionReceipt(DataTransferObject):
    """
    Receipt from a mined transaction.
    Args:
        blockHash:
        blockNumber:
        cumulativeGasUsed: total amount of gas used by block
        From:
        gasUsed: amount of gas used by this specific transaction
        logs:
        logsBloom:
        status: 1 if transaction succeeded, 0 otherwise.
        transactionHash:
        transactionIndex:
        to: Account to which this transaction was sent. If the transaction was a contract creation this value is set to None.
        contractAddress: Contract Account address created, f the transaction was a contract creation, or None otherwise.
    """

    def __init__(self, blockHash: hex, blockNumber: int, cumulativeGasUsed: int, From: str, gasUsed: int,
                 logsBloom: hex, status: int, transactionHash: hex, transactionIndex: int,
                 logs: [Log] = None, to: str = None, contractAddress: str = None):
        self.blockHash = blockHash
        self.blockNumber = blockNumber
        self.cumulativeGasUsed = cumulativeGasUsed
        self.From = From
        self.gasUsed = gasUsed
        self.logsBloom = logsBloom
        self.status = status
        self.transactionHash = transactionHash
        self.transactionIndex = transactionIndex
        self.logs = logs
        self.to = to
        self.contractAddress = contractAddress


class Account(DataTransferObject):
    """
    An Ethereum account.
    Args:
        address: Account address. Derived from public key.
        chain_id: ID of the chain the account is used in.
        secret: Account private key. A 256 bit number.
        domain: ENS Domain name. ie. niceguy.eth
    """

    def __init__(self, address: str, chain_id: int, secret: int = None, domain: str = None):
        self.address = address
        self.chain_id = chain_id
        self.secret = secret
        self.domain = domain

    def __str__(self):
        return self.address



"""
Load libin3 shared library for the current system, map function ABI, sets in3 network transport functions.

Example of RPC to In3-Core library, In3 Network and back.
```
+----------------+                               +----------+                       +------------+                        +------------------+
|                | in3.client.eth.block_number() |          |     in3_client_rpc    |            |  In3 Network Request   |                  |e
|     python     +------------------------------>+  python  +----------------------->   libin3   +------------------------>     python       |
|   application  |                               |   in3    |                       |  in3-core  |                        |  http_transport  |
|                <-------------------------------+          <-----------------------+            <------------------------+                  |
+----------------+     primitive or Object       +----------+     ctype object      +------------+  in3_req_add_response  +------------------+
```
"""
import ctypes as c
import platform

from pathlib import Path


def libin3_new(chain_id: int, transport: c.CFUNCTYPE, debug=False) -> int:
    """
    Instantiate new In3 Client instance.
    Args:
        chain_id (int): Chain id as integer
        transport: Transport function for the in3 network requests
        debug: Turn on debugger logging
    Returns:
         instance (int): Memory address of the client instance, return value from libin3_new
    """
    assert isinstance(chain_id, int)
    global libin3
    _map_function_signatures()
    # define this function as the transport for in3 requests from client to server and back
    libin3.in3_set_default_transport(transport)
    # register transport and verifiers (needed only once)
    libin3.in3_register_eth_full()
    # TODO: IPFS libin3.in3_register_ipfs();
    libin3.in3_register_eth_api()
    # TODO: in3_set_storage_handler(c, storage_get_item, storage_set_item, storage_clear, NULL);
    # enable logging
    if debug:
        # set logger level to TRACE
        libin3.in3_log_set_quiet_(False)
        libin3.in3_log_set_level_(0)
    return libin3.in3_for_chain_auto_init(chain_id)


def libin3_free(instance: int):
    """
    Free In3 Client objects from memory.
    Args:
        instance (int): Memory address of the client instance, return value from libin3_new
    """
    libin3.in3_free(instance)


def libin3_exec(instance: int, rpc: bytes):
    """
    Make Remote Procedure Call mapped methods in the client.
    Args:
        instance (int): Memory address of the client instance, return value from libin3_new
        rpc (bytes): Serialized function call, a json string.
    Returns:
        returned_value (object): The returned function value(s)
    """
    ptr_res = libin3.in3_client_exec_req(instance, rpc)
    result = c.cast(ptr_res, c.c_char_p).value
    libin3._free_(ptr_res)
    return result


def libin3_call(instance: int, fn_name: bytes, fn_args: bytes) -> (str, str):
    """
    Make Remote Procedure Call to an arbitrary method of a libin3 instance
    Args:
        instance (int): Memory address of the client instance, return value from libin3_new
        fn_name (bytes): Name of function that will be called in the client rpc.
        fn_args: (bytes) Serialized list of arguments, matching the parameters order of this function. i.e. ['0x123']
    Returns:
        result (int): Function execution status.
    """
    response = c.c_char_p()
    error = c.c_char_p()
    result = libin3.in3_client_rpc(instance, fn_name, fn_args, c.byref(response), c.byref(error))
    return result, response.value, error.value


def libin3_set_pk(instance: int, private_key: bytes):
    """
    Register the signer module in the In3 Client instance, with selected private key loaded in memory.
    Args:
        instance (int): Memory address of the client instance, return value from libin3_new
        private_key: 256 bit number.
    """
    libin3.eth_set_pk_signer_hex(instance, private_key)


def _multi_platform_selector(prefix: str) -> str:
    """
    Helper to define the path of installed shared libraries.
    Returns:
        libin3_file_path (pathlib.Path): Path to the correct library, compiled to the current platform.
    """
    path = Path(__file__).parent
    system, node, release, version, machine, processor = platform.uname()

    def fail():
        raise OSError('Not available on this platform ({}, {}, {}).'.format(
            system, processor, machine))

    # Fail over
    if not processor:
        processor = 'i386'
    # Similar behavior could be achieved with regex expressions if we known them better.
    suffix = None
    if processor == 'i386' or 'Intel' in processor:
        # AMD64 x86_64 64bit ...
        if '64' in machine:
            if system == 'Windows':
                suffix = "x64.dll"
            elif system == "Linux":
                suffix = "x64.so"
            elif system == 'Darwin':
                suffix = "x64.dylib"
        elif '32' in machine or '86' in machine:
            if system == 'Windows':
                fail()
            elif system == "Linux":
                suffix = "x86.so"
            elif system == 'Darwin':
                fail()
        elif 'armv' in machine:
            suffix = 'arm7.so'
    elif 'ARM' in processor:
        if machine == 'ARM7':
            suffix = 'arm7.so'
    if not suffix:
        fail()
    return str(Path(path.parent, "libin3", "shared", "{}.{}".format(prefix, suffix)))


def _map_function_signatures():
    # =================== LIBIN3 SHARED LIBRARY MAPPING ===================
    global libin3
    # map new in3
    libin3.in3_for_chain_auto_init.argtypes = [c.c_int]
    libin3.in3_for_chain_auto_init.restype = c.c_void_p
    libin3.in3_for_chain_default.argtypes = [c.c_int]
    libin3.in3_for_chain_default.restype = c.c_void_p
    # map free in3
    libin3.in3_free.argtypes = c.c_void_p,
    libin3.in3_free.restype = None
    libin3._free_.argtypes = c.c_void_p,
    libin3._free_.restype = None
    # map set pk signer
    libin3.eth_set_pk_signer_hex.argtypes = [c.c_void_p, c.c_char_p]
    libin3.eth_set_pk_signer_hex.restype = None
    # map transport request function
    libin3.in3_client_exec_req.argtypes = [c.c_void_p, c.c_char_p]
    libin3.in3_client_exec_req.restype = c.c_void_p
    libin3.in3_client_rpc.argtypes = [c.c_void_p, c.c_char_p, c.c_char_p,
                                      c.POINTER(c.c_char_p),
                                      c.POINTER(c.c_char_p)]
    libin3.in3_client_rpc.restype = c.c_int
    # map transport function for response
    libin3.in3_req_add_response.argtypes = [c.c_void_p, c.c_int, c.c_bool, c.c_char_p, c.c_int]
    libin3.in3_req_add_response.restype = None
    # map logging functions
    libin3.in3_log_set_quiet_.argtypes = c.c_bool,
    libin3.in3_log_set_quiet_.restype = None
    libin3.in3_log_set_level_.argtypes = c.c_int,
    libin3.in3_log_set_level_.restype = None


def init():
    """
    Loads library depending on host system.
    """
    return c.cdll.LoadLibrary(_multi_platform_selector('libin3'))


libin3 = init()



"""
Ethereum Domain Object Factory
For more on design-patterns see [Martin Fowler's](https://martinfowler.com/eaaCatalog/) Catalog of Patterns of Enterprise Application Architecture.
"""
from in3.eth.model import Block, Transaction, Log, TransactionReceipt, Account
from in3.exception import HashFormatException, EthAddressFormatException
from in3.libin3.enum import In3Methods
from in3.libin3.runtime import In3Runtime


class EthObjectFactory:
    """
    Deserialize and instantiate marshalled objects from rpc responses
    """

    def __init__(self, runtime: In3Runtime):
        self._runtime = runtime

    def _deserialize(self, obj_dict: dict, mapping: dict) -> dict:
        result = {}
        for m in mapping:
            if obj_dict.get(m) is None:
                result[m] = None
                continue
            result[m] = mapping[m](obj_dict[m])
        return result

    def get_integer(self, hex_str: str) -> int:
        if hex_str == "0x":
            return 0
        return int(hex_str, 16)

    def get_block(self, serialized: dict):
        mapping = {
            "number": self.get_integer,
            "hash": self.get_hash,
            "parentHash": self.get_hash,
            "nonce": self.get_integer,
            "sha3Uncles": str,
            "logsBloom": str,
            "transactionsRoot": self.get_hash,
            "stateRoot": self.get_hash,
            "miner": str,
            "author": str,
            "difficulty": self.get_integer,
            "totalDifficulty": self.get_integer,
            "extraData": str,
            "size": self.get_integer,
            "gasLimit": self.get_integer,
            "gasUsed": self.get_integer,
            "timestamp": self.get_integer,
            # "transactions": [{...}, {...}],
            # "uncles": list
        }
        obj = self._deserialize(serialized, mapping)
        transactions = []
        if serialized["transactions"] is not None and len(serialized["transactions"]) > 0:
            for t in serialized["transactions"]:
                if type(t) is dict:
                    aux = self.get_transaction(t)
                    transactions.append(aux)
                else:
                    # In case the user wants tx hashes only not full txs
                    transactions.append(self.get_hash(t))
        obj["transactions"] = transactions
        uncles = []
        if serialized["uncles"] is not None and len(serialized["uncles"]) > 0:
            for t in serialized["uncles"]:
                if type(t) is dict:
                    aux = self.get_block(t)
                    uncles.append(aux)
                else:
                    # In case the user wants block hashes only not full blocks
                    uncles.append(self.get_hash(t))
        obj["uncles"] = uncles
        return Block(**obj)

    def get_transaction(self, serialized: dict) -> Transaction:
        mapping = {
            "blockHash": self.get_hash,
            # "from": self.get_address,
            "gas": self.get_integer,
            "gasPrice": self.get_integer,
            "hash": self.get_hash,
            "input": str,
            "nonce": self.get_integer,
            "to": self.get_address,
            "transactionIndex": self.get_integer,
            "value": self.get_integer,
            "raw": str,
            "standardV": self.get_integer,
            "publicKey": str,
            "creates": str,
            "chainId": self.get_integer,
            "v": self.get_integer,
            "r": self.get_integer,
            "s": self.get_integer
        }
        aux = self._deserialize(serialized, mapping)
        aux["From"] = self.get_address(serialized["from"])
        return Transaction(**aux)

    def get_tx_receipt(self, serialized: dict) -> TransactionReceipt:
        mapping = {
            'blockHash': self.get_hash,
            'blockNumber': self.get_integer,
            'contractAddress': self.get_address if serialized['contractAddress'] else None,
            'cumulativeGasUsed': self.get_integer,
            'gasUsed': self.get_integer,
            'logsBloom': str,
            'status': self.get_integer,
            'to': self.get_address if serialized['to'] else None,
            'transactionHash': self.get_hash,
            'transactionIndex': self.get_integer
        }
        obj = self._deserialize(serialized, mapping)
        obj["From"] = self.get_address(serialized["from"])
        if serialized["logs"] is not None and len(serialized["logs"]) > 0:
            obj["logs"] = [self.get_log(log) for log in serialized["logs"]]
        return TransactionReceipt(**obj)

    def get_log(self, serialized: dict) -> Log:
        mapping = {
            'address': self.get_address,
            'blockHash': self.get_integer,
            'blockNumber': self.get_integer,
            'data': str,
            'logIndex': self.get_integer,
            'removed': bool,
            'transactionHash': self.get_hash,
            'transactionIndex': self.get_integer,
            'transactionLogIndex': self.get_integer,
        }
        aux = self._deserialize(serialized, mapping)
        aux["topics"] = [self.get_hash(topic) for topic in serialized["topics"]]
        aux["Type"] = serialized["type"]
        return Log(**aux)

    def checksum_address(self, address: str, add_chain_id: bool = True) -> str:
        return self._runtime.call(In3Methods.CHECKSUM_ADDRESS, address, add_chain_id)

    def get_hash(self, hash_str: str) -> str:
        if not hash_str.startswith('0x'):
            raise HashFormatException("Ethereum hashes start with 0x")
        if len(hash_str[2:].encode('utf-8')) != 64:
            raise HashFormatException("Hash size is not of an Ethereum hash.")
        return hash_str

    def get_address(self, address: str) -> str:
        if not address.startswith("0x"):
            raise EthAddressFormatException("Ethereum addresses start with 0x")
        if len(address.encode("utf-8")) != 42:
            raise EthAddressFormatException("The string don't have the size of an Ethereum address.")
        if self.checksum_address(address, False) != address:
            address = self.checksum_address(address, False)
        return address.replace('\'', '"')

    def get_account(self, address: str, secret: int = None) -> Account:
        return Account(self.get_address(address), self._runtime.chain_id, secret)



import random

from in3.eth.factory import EthObjectFactory
from in3.eth.model import NewTransaction, TransactionReceipt, Account
from in3.exception import PrivateKeyNotFoundException
from in3.libin3.enum import EthMethods, BlockAt, In3Methods
from in3.libin3.runtime import In3Runtime


class EthAccountApi:
    """
    Manages accounts and smart-contracts
    """

    def __init__(self, runtime: In3Runtime, factory: EthObjectFactory):
        self._runtime = runtime
        self._factory = factory

    def _create_account(self, secret: str) -> Account:
        if not secret.startswith('0x') or not len(secret) == 66:
            TypeError('Secret must be a 256 bit hexadecimal.')
        address = self._runtime.call(In3Methods.PK_2_ADDRESS, secret)
        account = self._factory.get_account(address, int(secret, 16))
        return account

    def create(self, qrng=False) -> Account:
        """
        Creates a new Ethereum account and saves it in the wallet.
        Args:
            qrng (bool): True uses a Quantum Random Number Generator api for generating the private key.
        Returns:
            account (Account): Newly created Ethereum account.
        """
        if qrng:
            raise NotImplementedError
        secret = hex(random.getrandbits(256))
        return self._create_account(secret)

    def recover(self, secret: str) -> Account:
        """
        Recovers an account from a secret.
        Args:
            secret (str): Account private key in hexadecimal string
        Returns:
            account (Account): Recovered Ethereum account.
        """
        if not secret or not len(secret) == 66:
            raise PrivateKeyNotFoundException('Please provide account secret.')
        return self._create_account(secret)

    def parse_mnemonics(self, mnemonics: str) -> str:
        """
        Recovers an account secret from mnemonics phrase
        Args:
            mnemonics (str): BIP39 mnemonics phrase.
        Returns:
            secret (str): Account secret. Use `recover_account` to create a new account with this secret.
        """
        raise NotImplementedError

    def sign(self, private_key: str, message: str) -> str:
        """
        Use ECDSA to sign a message.
        Args:
            private_key (str): Must be either an address(20 byte) or an raw private key (32 byte)"}}'
            message (str): Data to be hashed and signed. Dont input hashed data unless you know what you are doing.
        Returns:
            signed_message (str): ECDSA calculated r, s, and parity v, concatenated. v = 27 + (r % 2)
        """
        #   SIGN_EC_RAW  = 0, /**< sign the data directly
        #   SIGN_EC_HASH = 1, /**< hash and sign the data */
        signature_type = 'eth_sign'
        # in3_ret_t in3_sign_data(data, pk, sig_type)
        signature_dict = self._runtime.call(EthMethods.SIGN, message, private_key, signature_type)
        return signature_dict['signature']

    def balance(self, address: str, at_block: int or str = str(BlockAt.LATEST)) -> int:
        """
        Returns the balance of the account of given address.
        Args:
            address (str): address to check for balance
            at_block (int or str):  block number IN3BlockNumber  or EnumBlockStatus
        Returns:
            balance (int): integer of the current balance in wei.
        """
        account = self._factory.get_account(address)
        if isinstance(at_block, int):
            at_block = hex(at_block)
        return self._factory.get_integer(self._runtime.call(EthMethods.BALANCE, account.address, at_block))

    def send_transaction(self, sender: Account, transaction: NewTransaction) -> str:
        """
        Signs and sends the assigned transaction. Requires `account.secret` value set.
        Transactions change the state of an account, just the balance, or additionally, the storage and the code.
        Every transaction has a cost, gas, paid in Wei. The transaction gas is calculated over estimated gas times the
        gas cost, plus an additional miner fee, if the sender wants to be sure that the transaction will be mined in the
        latest block.
        Args:
            sender (Account): Sender Ethereum account. Senders generally pay the gas costs, so they must have enough balance to pay gas + amount sent, if any.
            transaction (NewTransaction): All information needed to perform a transaction. Minimum is to and value. Client will add the other required fields, gas and chaindId.
        Returns:
            tx_hash (hex): Transaction hash, used to get the receipt and check if the transaction was mined.
        """
        assert isinstance(transaction, NewTransaction)
        assert isinstance(sender, Account)
        if not sender.secret or not len(hex(sender.secret)) == 66:
            raise AssertionError('To send a transaction, the sender\'s secret must be known by the application. \
            To send a pre-signed transaction use `send_raw_transaction` instead.')
        transaction.From = sender.address
        self._runtime.set_signer_account(sender.secret)
        return self._runtime.call(EthMethods.SEND_TRANSACTION, transaction.serialize())

    def send_raw_transaction(self, signed_transaction: str) -> str:
        """
        Sends a signed and encoded transaction.
        Args:
            signed_transaction: Signed keccak hash of the serialized transaction
            Client will add the other required fields, gas and chaindId.
        Returns:
            tx_hash (hex): Transaction hash, used to get the receipt and check if the transaction was mined.
        """
        return self._runtime.call(EthMethods.SEND_RAW_TRANSACTION, signed_transaction)

    def estimate_gas(self, transaction: NewTransaction) -> int:
        """
        Gas estimation for transaction. Used to fill transaction.gas field. Check RawTransaction docs for more on gas.
        Args:
            transaction: Unsent transaction to be estimated. Important that the fields data or/and value are filled in.
        Returns:
            gas (int): Calculated gas in Wei.
        """
        gas = self._runtime.call(EthMethods.ESTIMATE_TRANSACTION, transaction.serialize())
        return self._factory.get_integer(gas)

    def transaction_count(self, address: str, at_block: int or str = str(BlockAt.LATEST)) -> int:
        """
        Number of transactions mined from this address. Used to set transaction nonce.
        Nonce is a value that will make a transaction fail in case it is different from (transaction count + 1).
        It exists to mitigate replay attacks.
        Args:
            address (str): Ethereum account address
            at_block (int):  Block number
        Returns:
            tx_count (int): Number of transactions mined from this address.
        """
        account = self._factory.get_account(address)
        if isinstance(at_block, int):
            at_block = hex(at_block)
        tx_count = self._runtime.call(EthMethods.TRANSACTION_COUNT, account.address, at_block)
        return self._factory.get_integer(tx_count)

    def checksum_address(self, address: str, add_chain_id: bool = True) -> str:
        """
        Will convert an upper or lowercase Ethereum address to a checksum address, that uses case to encode values.
        See [EIP55](https://github.com/ethereum/EIPs/blob/master/EIPS/eip-55.md).
        Args:
            address: Ethereum address string or object.
            add_chain_id (bool): Will append the chain id of the address, for multi-chain support, canonical for Eth.
        Returns:
            checksum_address: EIP-55 compliant, mixed-case address object.
        """
        return self._factory.checksum_address(self._factory.get_account(address).address, add_chain_id)



import ctypes as c

import requests

from in3.libin3.lib_loader import libin3


class In3Request(c.Structure):
    """
    Request sent by the libin3 to the In3 Network, transported over the _http_transport function
    Based on in3/client/.h in3_request_t struct
    Attributes:
        payload (str): the payload to send
        urls ([str]): array of urls
        urls_len (int): number of urls
        results (str): the responses
        timeout (int): the timeout 0= no timeout
        times (int): measured times (in ms) which will be used for ajusting the weights
    """
    _fields_ = [("payload", c.POINTER(c.c_char)),
                ("urls", c.POINTER(c.POINTER(c.c_char))),
                ("urls_len", c.c_int),
                ("results", c.c_void_p),
                ("timeout", c.c_uint32),
                ("times", c.c_uint32)]


def _transport_report_success(in3_request: In3Request, i: int, response: requests.Response):
    libin3.in3_req_add_response(in3_request.results, i, False, response.content, len(response.content))


def _transport_report_failure(in3_request: In3Request, i: int, msg: bytes):
    libin3.in3_req_add_response(in3_request.results, i, True, msg, len(msg))


@c.CFUNCTYPE(c.c_int, c.POINTER(In3Request))
def http_transport(in3_request: In3Request):
    """
    Transports each request coming from libin3 to the in3 network and and reports the answer back
    Args:
        in3_request (In3Request): request sent by the In3 Client Core to the In3 Network
    Returns:
        exit_status (int): Always zero for signaling libin3 the function executed OK.
    """
    in3_request = in3_request.contents

    for i in range(0, in3_request.urls_len):
        try:
            http_request = {
                'url': c.string_at(in3_request.urls[i]),
                'data': c.string_at(in3_request.payload),
                'headers': {'Content-type': 'application/json'},
                'timeout': in3_request.timeout
            }
            response = requests.post(**http_request)
            response.raise_for_status()
            if 'error' in response.text:
                _transport_report_failure(in3_request, i, response.content)
            else:
                _transport_report_success(in3_request, i, response)
        except requests.exceptions.RequestException as err:
            _transport_report_failure(in3_request, i, str(err).encode('utf8'))
        except requests.exceptions.SSLError as err:
            _transport_report_failure(in3_request, i, str(err).encode('utf8'))
        except Exception as err:
            _transport_report_failure(in3_request, i, str(err).encode('utf8'))
    return 0



import random

from in3.eth.factory import EthObjectFactory
from in3.eth.model import Account
from in3.libin3.enum import In3Methods
from in3.libin3.runtime import In3Runtime


class WalletApi:
    """
    Ethereum accounts holder and manager.
    """

    def __init__(self, runtime: In3Runtime):
        self._runtime = runtime
        self._factory = EthObjectFactory(runtime)
        # TODO: Persistence
        self._accounts: dict = {}

    def new_account(self, name: str, qrng=False) -> Account:
        """
        Creates a new Ethereum account and saves it in the wallet.
        Args:
            name (str): Account identifier to use with `get`. i.e. get('my_wallet`)
            qrng (bool): True uses a quantum random number generator api for generating the private key.
        Returns:
            account (Account): Newly created Ethereum account.
        """
        if qrng:
            raise NotImplementedError
        secret = hex(random.getrandbits(256))
        return self._create_account(name, secret)

    def get(self, name: str) -> Account or None:
        """
        Returns account in case it exists. Doesnt fail for easier handling w/out try catch. i.e: if get('my_wallet'): do
        Args:
            name (str): Account identifier to use with `get`. i.e. get('my_wallet`)
        Returns:
            account (Account): Selected Ethereum account.
        """
        if name not in self._accounts.keys():
            return None
        return self._accounts[name]

    def delete(self, name: str) -> bool:
        """
        Deletes an account in case it exists.
        Args:
            name (str): Account identifier.
        Returns:
            success (bool): True if account exists and was deleted.
        """
        if name not in self._accounts.keys():
            return False
        del self._accounts[name]
        return True

    def recover_account(self, name: str, secret: str) -> Account:
        """
        Recovers an account from a secret.
        Args:
            name (str): Account identifier to use with `get`. i.e. get('my_wallet`)
            secret (str): Account private key in hexadecimal string
        Returns:
            account (Account): Recovered Ethereum account.
        """
        return self._create_account(name, secret)

    def parse_mnemonics(self, mnemonics: str) -> str:
        """
        Recovers an account secret from mnemonics phrase
        Args:
            mnemonics (str): BIP39 mnemonics phrase.
        Returns:
            secret (str): Account secret. Use `recover_account` to create a new account with this secret.
        """
        raise NotImplementedError

    def _create_account(self, name: str, secret: str) -> Account:
        if not secret.startswith('0x') or not len(secret) == 66:
            TypeError('Secret must be a 256 bit hexadecimal.')
        address = self._runtime.call(In3Methods.PK_2_ADDRESS, secret)
        account = self._factory.get_account(address, int(secret, 16))
        self._accounts[name] = account
        return account



import in3.eth.model as eth
from in3.client import Client
from in3.exception import EthAddressFormatException, HashFormatException, PrivateKeyNotFoundException, ClientException
from in3.model import ClientConfig, NodeList

__name__ = 'in3'
__author__ = 'Slock.it <github.com/slockit>'
__repository__ = 'github.com/slockit/in3-c'
__status__ = "beta"
__version__ = "2.1"
__date__ = "May 2020"



from in3.eth.api import EthereumApi
from in3.eth.factory import EthObjectFactory
from in3.libin3.enum import In3Methods
from in3.libin3.runtime import In3Runtime
from in3.model import In3Node, NodeList, ClientConfig, ChainConfig, chain_configs
from in3.transport import http_transport


class Client:
    """
    Incubed network client. Connect to the blockchain via a list of bootnodes, then gets the latest list of nodes in
    the network and ask a certain number of the to sign the block header of given list, putting their deposit at stake.
    Once with the latest list at hand, the client can request any other on-chain information using the same scheme.
    Args:
        in3_config (ClientConfig or str): (optional) Configuration for the client. If not provided, default is loaded.
    """

    def __init__(self, chain: str or ChainConfig = 'mainnet',
                 in3_config: ClientConfig = None, transport=http_transport):

        config = in3_config
        if isinstance(chain, ChainConfig):
            config = chain.client_config
        elif not isinstance(chain, str) or chain not in ['mainnet', 'kovan', 'goerli']:
            raise ValueError('Chain name not supported. Try mainnet, kovan, goerli.')

        self._runtime = In3Runtime(chain_configs[chain].chain_id, transport)
        if config:
            # TODO: Chain_configs
            self._configure(config)
        # TODO: getConfig
        self.eth = EthereumApi(self._runtime)
        self._factory = In3ObjectFactory(self._runtime)

    def _configure(self, in3_config: ClientConfig) -> bool:
        """
        Send RPC to change client configuration. Don't use outside the constructor, might cause instability.
        """
        fn_args = str([in3_config.serialize()]).replace('\'', '')
        return self._runtime.call(In3Methods.CONFIGURE, fn_args, formatted=True)

    def refresh_node_list(self) -> NodeList:
        """
        Gets the list of Incubed nodes registered in the selected chain registry contract.
        Returns:
            node_list (NodeList): List of registered in3 nodes and metadata.
        """
        node_list_dict = self._runtime.call(In3Methods.NODE_LIST)
        return self._factory.get_node_list(node_list_dict)

    def config(self) -> dict:
        # TODO: Marshalling
        return self._runtime.call(In3Methods.GET_CONFIG)

    def ens_namehash(self, domain_name: str) -> str:
        """
        Name format based on [EIP-137](https://github.com/ethereum/EIPs/blob/master/EIPS/eip-137.md#name-syntax)
        Args:
            domain_name: ENS supported domain. mydomain.ens, mydomain.xyz, etc
        Returns:
            node (str): Formatted string referred as `node` in ENS documentation
        """
        return self._runtime.call(In3Methods.ENSRESOLVE, domain_name, 'hash')

    def ens_address(self, domain_name: str, registry: str = None) -> str:
        """
        Resolves ENS domain name to what account that domain points to.
        Args:
            domain_name: ENS supported domain. mydomain.ens, mydomain.xyz, etc
            registry: ENS registry contract address. i.e. 0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e
        Returns:
            address (str): Ethereum address corresponding to what account that domain points to.
        """
        # TODO: Add handlers to Account
        # TODO: Add serialization in account factory
        if registry:
            registry = self._factory.get_address(registry)
        return self._runtime.call(In3Methods.ENSRESOLVE, domain_name, 'addr', registry)

    def ens_owner(self, domain_name: str, registry: str = None) -> str:
        """
        Resolves ENS domain name to Ethereum address of domain owner.
        Args:
            domain_name: ENS supported domain. mydomain.ens, mydomain.xyz, etc
            registry: ENS registry contract address. i.e. 0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e
        Returns:
            owner_address (str): Ethereum address corresponding to domain owner.
        """
        if registry:
            registry = self._factory.get_address(registry)
        return self._runtime.call(In3Methods.ENSRESOLVE, domain_name, 'owner', registry)

    def ens_resolver(self, domain_name: str, registry: str = None) -> str:
        """
        Resolves ENS domain name to Smart-contract address of the resolver registered for that domain.
        Args:
            domain_name: ENS supported domain. mydomain.ens, mydomain.xyz, etc
            registry: ENS registry contract address. i.e. 0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e
        Returns:
            resolver_contract_address (str): Smart-contract address of the resolver registered for that domain.
        """
        if registry:
            registry = self._factory.get_address(registry)
        return self._runtime.call(In3Methods.ENSRESOLVE, domain_name, 'resolver', registry)


class In3ObjectFactory(EthObjectFactory):

    # TODO: Decode abi for props
    # TODO: Convert registerTime to human readable
    def get_node_list(self, serialized: dict) -> NodeList:
        mapping = {
            "nodes": list,
            "contract": self.get_account,
            "registryId": str,
            "lastBlockNumber": int,
            "totalServers": int,
        }
        deserialized_dict = self._deserialize(serialized, mapping)
        nodes = [self.get_in3node(node) for node in deserialized_dict['nodes']]
        deserialized_dict['nodes'] = nodes
        return NodeList(**deserialized_dict)

    def get_in3node(self, serialized: dict) -> In3Node:
        mapping = {
            "url": str,
            "address": self.get_account,
            "index": int,
            "deposit": self.get_integer,
            "props": str,
            "timeout": int,
            "registerTime": int,
            "weight": int
        }
        deserialized_dict = self._deserialize(serialized, mapping)
        return In3Node(**deserialized_dict)



"""
Integrated tests for `in3`

Tests data marshalling from and to libin3.
Tests minimum, happy path, client consistency and stability.
Calls all implemented calls from Python to libin3.
Uses real network data mocked locally.
"""



"""
Integrated tests for `in3` module. Doesnt test submodules.
"""
import unittest

import in3
from tests.mock.config import mock_config
from tests.transport import mock_transport


class MainNetClientTest(unittest.TestCase):

    def setUp(self):
        # self.client = in3.Client(in3_config=mock_config)
        self.client = in3.Client(in3_config=mock_config, transport=mock_transport)

    def test_configure(self):
        client = in3.Client()
        self.assertIsInstance(client, in3.Client)
        client = in3.Client(in3_config=in3.model.ClientConfig())
        self.assertIsInstance(client, in3.Client)

    def test_node_list(self):
        node_list = self.client.refresh_node_list()
        self.assertIsInstance(node_list, in3.model.NodeList)

    def test_ens_resolve(self):
        # Other calls like `addr` require more than one eth_call, being more complex to mock the tests. Suffice for now.
        with self.assertRaises(in3.EthAddressFormatException):
            self.client.ens_owner('depraz.eth', 'asd')
        address = self.client.ens_owner('depraz.eth')
        self.assertEqual(address, '0x0b56ae81586d2728ceaf7c00a6020c5d63f02308')

    def test_ens_namehash(self):
        foo_name = self.client.ens_namehash('foo.eth')
        self.assertEqual(foo_name, '0xde9b09fd7c5f901e23a3f19fecc54828e9c848539801e86591bd9801b019f84f')


class KovanClientTest(MainNetClientTest):

    def setUp(self):
        # self.client = in3.Client('kovan', in3_config=mock_config)
        self.client = in3.Client('kovan', in3_config=mock_config, transport=mock_transport)

    def test_configure(self):
        client = in3.Client('kovan')
        self.assertIsInstance(client, in3.Client)
        client = in3.Client('kovan', in3.model.ClientConfig())
        self.assertIsInstance(client, in3.Client)

    def test_ens_resolve(self):
        # Not supported by app.ens.domains!
        return


class GoerliClientTest(MainNetClientTest):

    def setUp(self):
        # self.client = in3.Client('goerli', in3_config=mock_config)
        self.client = in3.Client('goerli', in3_config=mock_config, transport=mock_transport)

    def test_configure(self):
        client = in3.Client('goerli')
        self.assertIsInstance(client, in3.Client)
        client = in3.Client('goerli', in3.model.ClientConfig())
        self.assertIsInstance(client, in3.Client)

    def test_ens_resolve(self):
        # Other calls like `addr` require more than one eth_call, being more complex to mock the tests. Suffice for now.
        address = self.client.ens_owner('depraz.eth')
        self.assertEqual(address, '0x0b56ae81586d2728ceaf7c00a6020c5d63f02308')


if __name__ == '__main__':
    unittest.main()



import re

from in3.eth.factory import EthObjectFactory
from in3.eth.model import NewTransaction, DataTransferObject
from in3.libin3.enum import EthMethods, BlockAt, In3Methods
from in3.libin3.runtime import In3Runtime


class SmartContract(DataTransferObject):
    """
    Ethereum account containing smart-contract code.
    Args:
        address: Account address. Derived from public key.
        chain_id: ID of the chain the account is used in.
        abi: Contract methods ABI, used to call them via RPC. Parsed from json string.
        code: EVM Bytecode for contract execution. Important for STANDARD and FULL verification.
        secret: Account private key. A 256 bit number.
        domain: ENS Domain name. ie. niceguy.eth
    """

    def __init__(self, address: str, chain_id: int, abi: dict, code: str = None, secret: int = None,
                 domain: str = None):
        self.address = address
        self.chain_id = chain_id
        self.abi = abi
        self.code = code
        self.secret = secret
        self.domain = domain

    def __str__(self):
        return self.address


class EthContractApi:
    """
    Manages smart-contract data and transactions
    """

    def __init__(self, runtime: In3Runtime, factory: EthObjectFactory):
        self._runtime = runtime
        self._factory = factory

    def call(self, transaction: NewTransaction, block_number: int or str = 'latest') -> int or str:
        """
        Calls a smart-contract method. Will be executed locally by Incubed's EVM or signed and sent over to save the state changes.
        Check https://ethereum.stackexchange.com/questions/3514/how-to-call-a-contract-method-using-the-eth-call-json-rpc-api for more.
        Args:
            transaction (NewTransaction):
            block_number (int or str):  Desired block number integer or 'latest', 'earliest', 'pending'.
        Returns:
            method_returned_value: A hexadecimal. For decoding use in3.abi_decode.
        """
        # different than eth_call
        # eth_call_fn(c, contract, BLKNUM_LATEST(), "servers(uint256):(string,address,uint,uint,uint,address)", to_uint256(i));
        return self._runtime.call(EthMethods.CALL, transaction.serialize(), block_number)

    def storage_at(self, address: str, position: int = 0, at_block: int or str = str(BlockAt.LATEST)) -> str:
        """
        Stored value in designed position at a given address. Storage can be used to store a smart contract state, constructor or just any data.
        Each contract consists of a EVM bytecode handling the execution and a storage to save the state of the contract.
        The storage is essentially a key/value store. Use get_code to get the smart-contract code.
        Args:
            address (str): Ethereum account address
            position (int):  Position index, 0x0 up to 0x64
            at_block (int or str):  Block number
        Returns:
            storage_at (str): Stored value in designed position. Use decode('hex') to see ascii format of the hex data.
        """
        account = self._factory.get_account(address)
        if isinstance(at_block, int):
            at_block = hex(at_block)
        return self._runtime.call(EthMethods.STORAGE_AT, account.address, hex(position), at_block)

    def code(self, address: str, at_block: int or str = str(BlockAt.LATEST)) -> str:
        """
        Smart-Contract bytecode in hexadecimal. If the account is a simple wallet the function will return '0x'.
        Args:
            address (str): Ethereum account address
            at_block (int or str): Block number
        Returns:
            bytecode (str): Smart-Contract bytecode in hexadecimal.
        """
        account = self._factory.get_account(address)
        if isinstance(at_block, int):
            at_block = hex(at_block)
        return self._runtime.call(EthMethods.CODE, account.address, at_block)

    def encode(self, fn_signature: str, *fn_args) -> str:
        """
        Smart-contract ABI encoder. Used to serialize a rpc to the EVM.
        Based on the [Solidity specification.](https://solidity.readthedocs.io/en/v0.5.3/abi-spec.html)
        Note: Parameters refers to the list of variables in a method declaration.
        Arguments are the actual values that are passed in when the method is invoked.
        When you invoke a method, the arguments used must match the declaration's parameters in type and order.
        Args:
            fn_signature (str): Function name, with parameters. i.e. `getBalance(uint256):uint256`, can contain the return types but will be ignored.
            fn_args (tuple): Function parameters, in the same order as in passed on to method_name.
        Returns:
            encoded_fn_call (str): i.e. "0xf8b2cb4f0000000000000000000000001234567890123456789012345678901234567890"
        """
        self._check_fn_signature(fn_signature)
        return self._runtime.call(In3Methods.ABI_ENCODE, fn_signature, fn_args)

    def decode(self, fn_signature: str, encoded_value: str) -> tuple:
        """
        Smart-contract ABI decoder. Used to parse rpc responses from the EVM.
        Based on the [Solidity specification.](https://solidity.readthedocs.io/en/v0.5.3/abi-spec.html)
        Args:
            fn_signature: Function signature. e.g. `(address,string,uint256)` or `getBalance(address):uint256`.
            In case of the latter, the function signature will be ignored and only the return types will be parsed.
            encoded_value: Abi encoded values. Usually the string returned from a rpc to the EVM.
        Returns:
            decoded_return_values (tuple):  "0x1234567890123456789012345678901234567890", "0x05"
        """
        if not encoded_value.startswith('0x'):
            raise AssertionError("Encoded values must start with 0x")
        if len(encoded_value[2:]) < 64:
            raise AssertionError("Encoded values must be longer than 64 characters.")
        self._check_fn_signature(fn_signature)
        return self._runtime.call(In3Methods.ABI_DECODE, fn_signature, encoded_value)

    @staticmethod
    def _check_fn_signature(fn_signature):
        is_signature = re.match(r'.*(\(.+\))', fn_signature)
        _types = ["address", "string", "uint", "string", "bool", "bytes", "int"]
        contains_type = [_type for _type in _types if _type in fn_signature]
        if not is_signature or not contains_type:
            raise AssertionError('Function signature is not valid. A valid example is balanceOf(address).')



import pathlib
import setuptools
from os import environ

"""
Information on how to build this file
https://pypi.org/
https://pypi.org/classifiers/
"""


version = "v2.3.2-RC11"
url = environ.get("url", "https://github.com/slockit/in3-c")
License = environ.get("license", "AGPL")
description = environ.get(
    "description", "Incubed client and provider for web3. Based on in3-c runtime.")
keywords = environ.get(
    "keywords", "in3,c,arm,x86,x64,macos,windows,linux,blockchain,ethereum,bitcoin,ipfs").split(",")
readme = (pathlib.Path(__file__).parent / "README.md").read_text()
name = environ.get("name", "in3")
author = environ.get("author", "github.com/slockit/in3-c")
author_email = environ.get("author_email", "products@slock.it")
setuptools.setup(
    name=name,
    version=version,
    author=author,
    author_email=author_email,
    description=description,
    long_description=readme,
    long_description_content_type="text/markdown",
    url=url,
    packages=setuptools.find_packages(exclude=["docs", "tests"]),
    install_requires=['requests==2.23.0'],
    keywords=keywords,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
        "Environment :: Console",
        "Development Status :: 4 - Beta"
    ]
)
